//*************************************************************************
//  sLab-Car(エスラボ・カー)
//  Scratch3など遠隔制御用【標準モデル（カメラ、両サイド赤外線センサ有り）】
//*************************************************************************
// ①ライブラリの読み込み
#include <WiFi.h>
#include <Wire.h>
#include "ESPAsyncWebServer.h"
#include "WebSocketsServer.h"
#include "OV7670.h"

// ####################### ②Wi-Fi設定(環境設定) #######################
const char *ssid      = "##### SSID #####";
const char *password  = "### PASSWORD ###";
IPAddress ip(192, 168, 1, 123);  // IPアドレス（本機が利用するIP）
IPAddress gateway(192, 168, 1, 1);  // デフォルトゲートウェイ
IPAddress subnet(255, 255, 255, 0); // サブネットマスク
// ###################################################################

// ③ピン配置
const byte LED_PIN      =  2;  // LED緑
const byte R_WHEEL_IN1  = 16;  // 右モータ制御
const byte R_WHEEL_IN2  = 17;  // 右モータ制御
const byte L_WHEEL_IN1  = 18;  // 左モータ制御
const byte L_WHEEL_IN2  = 19;  // 左モータ制御
const byte ONPA_TRIG    =  4;  // 超音波トリガー
const byte ONPA_ECHO    =  5;  // 超音波エコー
const byte SERV_PIN     = 12;  // サーボモータ [GPIO12=A15]
const byte IR_LEFT_U    = 23;  // 上部、左赤外線センサ
const byte IR_RIGH_U    = 27;  // 上部、右赤外線センサ
const byte LEDC_CH      =  7;  // LEDC-CH

// ④カメラピン配置
const camera_config_t cam_conf = {
  .D0 = 36,
  .D1 = 39,
  .D2 = 34,
  .D3 = 35,
  .D4 = 32,
  .D5 = 33,
  .D6 = 25,
  .D7 = 26,
  .XCLK = 15,   // ★27利用時 Wi-Fi接続不可
  .PCLK = 14,
  .VSYNC = 13,
  .xclk_freq_hz = 10000000,     // XCLK 10MHz
  .ledc_timer   = LEDC_TIMER_0,
  .ledc_channel = LEDC_CHANNEL_0 
};

// ⑤カメラ解像度指定
#define CAM_RES     QCIF    // カメラ解像度
#define CAM_WIDTH   176     // カメラ幅
#define CAM_HEIGHT  144     // カメラ高さ
#define CAM_DIV       1     // １画面分割数

// ⑥グローバル変数宣言
OV7670 cam;                 // camera
AsyncWebServer server(80);  // WebサーバPort80番
WebSocketsServer webSocket = WebSocketsServer(81);
uint8_t clientNum;

bool wsFlag  = false;       // WebSocket状態フラグ(true:接続状態、false:未接続状態)
bool ledFlag = true;        // LED制御フラグ
int dataSize = 2 * CAM_WIDTH * CAM_HEIGHT;      // カメラデータ: 2 * 176 * 144 = 50688 [Byte]
uint8_t wsCamData[50688];   // カメラデータを保持
int adjustAngle    = 0;     // Angle補正
int preAdjustAngle = 0;     // 前回Angle補正
byte preAngle      = -1;    // 現在の角度を保存し、角度変更があるか把握
byte preCarMove    = 1; // 1:初期/Go、2:Left、3:Right、4:Back、5:Other

// ⑦Arduino初期設定処理
void setup() {
  // ⑧処理開始
  Serial.begin(115200);
  Serial.println("sLab-Car Sart"); 
  Wire.begin();
  Wire.setClock(400000);
  // ⑨端子設定
  pinMode ( LED_PIN, OUTPUT );
  pinMode(R_WHEEL_IN1, OUTPUT);
  pinMode(L_WHEEL_IN1, OUTPUT);
  pinMode(R_WHEEL_IN2, OUTPUT);
  pinMode(L_WHEEL_IN2, OUTPUT);
  pinMode(ONPA_TRIG, OUTPUT);
  pinMode(ONPA_ECHO, INPUT);
  pinMode(IR_LEFT_U, INPUT);
  pinMode(IR_RIGH_U, INPUT);
  ledcSetup(LEDC_CH, 50, 10);       // ch:0, Freq:50Hz, 10bit
  ledcAttachPin(SERV_PIN, LEDC_CH); // ch:0
  // ⑩端子の出力を初期値へ
  digitalWrite(R_WHEEL_IN1, LOW);
  digitalWrite(R_WHEEL_IN2, LOW);
  digitalWrite(L_WHEEL_IN1, LOW);
  digitalWrite(L_WHEEL_IN2, LOW);
  // ⑪サーボモータ回転(前向き[90度]へ調整) 
  ledcWrite(LEDC_CH, (76 + adjustAngle));

  // ⑫無線Wi-Fi接続
  WiFi.config( ip, gateway, subnet );
  WiFi.begin ( ssid, password );
  while ( WiFi.status() != WL_CONNECTED ) { // 接続するまで無限ループ処理
    // ⑬接続中はLEDを１秒毎に点滅する
    ledFlag = !ledFlag;
    digitalWrite(LED_PIN, ledFlag);
    delay ( 1000 );
    Serial.print ( "." );
  }
  // ⑭Wi-Fi接続完了（IPアドレス表示）
  Serial.print ( "Wi-Fi Connected! IP address: " );
  Serial.println ( WiFi.localIP() );
  // ⑮Wi-Fi接続時はLED点灯（Wi-Fi接続状態）
  digitalWrite ( LED_PIN, true );

  // ⑯カメラ初期化処理
  Serial.println( "cam init" );  
  esp_err_t err = cam.init(&cam_conf, CAM_RES, RGB565);
  if(err != ESP_OK){
    Serial.print( "cam.init ERR:" );
    Serial.println(err);
    // ⑰失敗時は15秒後に再起動しRetry
    delay(15000);
    ESP.restart();
  }
  cam.vflip( false );

  // ⑱WebServerアクセス処理設定[ホーム画面表示]
  server.on("/", HTTP_GET, [](AsyncWebServerRequest *request){
      sendTop(request);
      Serial.println ( "WEB TOP" );
  });
  // ⑲WebServerアクセス処理設定[LED制御]
  server.on("/led", HTTP_GET, [](AsyncWebServerRequest *request){
      String setLedVal = request->getParam("o")->value();
      if (setLedVal == "ON") {
        digitalWrite ( LED_PIN, true );
        Serial.println ( "led on" );
      } else {
        digitalWrite ( LED_PIN, false );
        Serial.println ( "led off" );
      }
      request->send(200, "text/plain", "OK");
  });
  // ⑳WebServerアクセス処理設定[超音波センサー角度補正]  
  server.on("/adjusta", HTTP_GET, [](AsyncWebServerRequest *request){
      adjustAngle = request->getParam("n")->value().toInt();
      Serial.println ( "angle adjest:" + String(adjustAngle) );
      request->send(200, "text/plain", "OK");
  });
  // ㉑WebServerアクセス処理設定[超音波センサー角度制御]  
  server.on("/s_angle", HTTP_GET, [](AsyncWebServerRequest *request){
      double angleNum = double(request->getParam("n")->value().toInt());
      Serial.println ( "servo angle:" + String(angleNum) );
      // ㉒現在の角度（アングル）から変更があるか判定
      if (preAngle != int(angleNum) || adjustAngle != preAdjustAngle) {
        // ㉓サーボモータ解説
        // パルス幅(ON時間)で角度が決まる（0度:0.5msec[=>26], 180度:2.4msec[=>123]）
        // 90度の場合) 180:1.9(2.4-0.5) = 90:x => x = 1.9*90/180 = 0.95
        // 0.95+0.5 = 1.45msec => 1.45 / 20(10kHz) *1024 => 75 
        // y度の場合) x = ( (1.9*y/180) + 0.5 ) / 20 * 1024
        int angleTime = ceil( ( (1.9*angleNum/180) + 0.5 ) / 20 * 1024 );
        angleTime = angleTime + adjustAngle;
        // 0度:0.5msec[=>26], 180度:2.4msec[=>123]なので、21から128までで制御
        if (angleTime < 21) {
          angleTime = 21;
        } else if (angleTime > 128) {
          angleTime = 128;          
        }
        // ㉔サーボモータ回転
        ledcWrite(LEDC_CH, angleTime);
        Serial.println ( "servo angle control:" + request->getParam("n")->value() + ":" + String(angleTime) );
        // ㉕角度を変更する時間を取って返答
        delay ( 1000 );
      // ㉖変更がないときはサーボモータ処理しない
      } else {
        Serial.println ( "servo angle noChange" );
      }
      // ㉗現在の角度を保持
      preAngle = int(angleNum);
      preAdjustAngle = adjustAngle;
      request->send(200, "text/plain", "OK");
  });
  // ㉘WebServerアクセス処理設定[超音波センサー距離測定]  
  server.on("/update_d", HTTP_GET, [](AsyncWebServerRequest *request){
      Serial.println ( "update distance" );
      // ㉙距離測定と送信メッセージ作成
      String sendMsg = updateSensor();
      request->send(200, "application/json", sendMsg);
  });
  // ㉚WebServerアクセス処理設定[Car前進、停止、後退制御]  
  server.on("/cargo", HTTP_GET, [](AsyncWebServerRequest *request){
      // ㉛要求内容を判断しカー制御
      if ( request->getParam("a")->value() == "STOP" ){
         Serial.println("STOP");
         car_stop();
      } else if ( request->getParam("a")->value() == "BACK" ){
         Serial.println("BACK");
         car_back();
      } else {
        Serial.println("GO");
        car_go();
      }
      request->send(200, "text/plain", "OK");
  });
  // ㉜WebServerアクセス処理設定[Car右左折制御]  
  server.on("/carturn", HTTP_GET, [](AsyncWebServerRequest *request){
      // ㉝要求内容を判断しカー制御
      if ( request->getParam("a")->value() == "RIGHT" ){
        Serial.println("RIGHT");
        car_right ( request->getParam("n")->value().toInt() );
      } else {
        Serial.println("LEFT");
        car_left ( request->getParam("n")->value().toInt() );
      }
      request->send(200, "text/plain", "OK");
  });
  // ㉞Responseヘッダオプション追加処理
  DefaultHeaders::Instance().addHeader("Access-Control-Allow-Origin", "*");
  // ㉟WebServer起動  
  server.begin();
  // ㊱WebSocket起動
  webSocket.begin();
  webSocket.onEvent(onWebSocketEvent);
  Serial.println("##### setup completed #####");
}

// ㊲Arduinoループ処理（通常時はこの処理）
void loop(void) {
  // ㊳WebSocket処理
  webSocket.loop();
  // ㊴WebSocket接続判定
  if (wsFlag) {
    // ㊵カメラデータ読み込み処理
    cam.getLines( 1 , wsCamData , CAM_HEIGHT);
    // ㊶映像データ送信処理
    webSocket.sendBIN(clientNum, wsCamData, dataSize);
  }
}
